"""
Import plugin that uses data files from 
https://elite.tromador.com/ to update the Database.
"""
from __future__ import annotations

from contextlib import contextmanager
from email.utils import parsedate_to_datetime
from pathlib import Path
import csv
import datetime
import os
import requests
import time
import typing

from ..fs import file_line_count
from .. import plugins, cache, transfers
from ..misc import progress as pbar
from ..plugins import PluginException

from sqlalchemy.orm import Session
from sqlalchemy import func, delete, select, exists, text

from tradedangerous import plugins, transfers
from tradedangerous.db import orm_models as SA, lifecycle
from tradedangerous.db.utils import (
    begin_bulk_mode, end_bulk_mode,
    get_import_batch_size, get_upsert_fn,
)
from tradedangerous.fs import file_line_count
from tradedangerous.misc import progress as pbar
from tradedangerous.plugins import PluginException

if typing.TYPE_CHECKING:
    from tradedangerous.tradeenv import TradeEnv

# Constants
BASE_URL = os.environ.get('TD_SERVER') or "https://elite.tromador.com/files/"


class DecodingError(PluginException):
    pass


@contextmanager
def bench(label: str, tdenv: TradeEnv):
    started = time.time()
    with pbar.Progress(1, 40, prefix=label):
        yield
    tdenv.NOTE("{} done ({:.3f}s)", label, time.time() - started)


def _count_listing_entries(tdenv: TradeEnv, listings: Path) -> int:
    """ Calculates the number of entries in a listing file by counting the lines. """
    if not listings.exists():
        tdenv.NOTE("File not found, aborting: {}", listings)
        return 0
    
    tdenv.DEBUG0(f"Getting total number of entries in {listings}...")
    count = file_line_count(listings)
    if count <= 1:
        if count == 1:
            tdenv.DEBUG0("Listing count of 1 suggests nothing but a header")
        else:
            tdenv.DEBUG0("Listings file is empty, nothing to do.")
        return 0
    
    return count + 1  # kfsone: Doesn't the header already make this + 1?


def _make_item_id_lookup(tdenv: TradeEnv, session: Session) -> frozenset[int]:
    """Helper: retrieve the list of commodities in database."""
    tdenv.DEBUG0("Getting list of commodities...")
    rows = session.query(SA.Item.item_id).all()
    return frozenset(r[0] for r in rows)


def _make_station_id_lookup(tdenv: TradeEnv, session: Session) -> frozenset[int]:
    """Helper: retrieve the list of station IDs in database."""
    tdenv.DEBUG0("Getting list of stations...")
    rows = session.query(SA.Station.station_id).all()
    return frozenset(r[0] for r in rows)


def _collect_station_modified_times(tdenv: TradeEnv, session: Session) -> dict[int, int]:
    """Helper: build a list of the last modified time for all stations by id (epoch seconds)."""
    tdenv.DEBUG0("Getting last-update times for stations...")
    rows = (
        session.query(
            SA.StationItem.station_id,
            func.min(SA.StationItem.modified),
        )
        .group_by(SA.StationItem.station_id)
        .all()
    )
    return {
        station_id: int(modified.timestamp()) if modified else 0
        for station_id, modified in rows
    }


class ImportPlugin(plugins.ImportPluginBase):
    """
    Import plugin that uses data files from
    https://elite.tromador.com/ to update the Database.
    """
    pluginOptions = {
        'item':         "Update Items using latest file from server. (Implies '-O system,station')",
        'rare':         "Update RareItems using latest file from server. (Implies '-O system,station')",
        'ship':         "Update Ships using latest file from server.",
        'upgrade':      "Update Upgrades using latest file from server.",
        'system':       "Update Systems using latest file from server.",
        'station':      "Update Stations using latest file from server. (Implies '-O system')",
        'shipvend':     "Update ShipVendors using latest file from server. (Implies '-O system,station,ship')",
        'upvend':       "Update UpgradeVendors using latest file from server. (Implies '-O system,station,upgrade')",
        'listings':     "Update market data using latest listings.csv dump. (Implies '-O item,system,station')",
        'all':          "Update everything with latest dumpfiles. (Regenerates all tables)",
        'clean':        "Erase entire database and rebuild from empty. (Regenerates all tables.)",
        'skipvend':     "Don't regenerate ShipVendors or UpgradeVendors. (Supercedes '-O all', '-O clean'.)",
        'force':        "Force regeneration of selected items even if source file not updated since previous run. "
                        "(Useful for updating Vendor tables if they were skipped during a '-O clean' run.)",
        'purge':        "Remove any empty systems that previously had fleet carriers.",
        'optimize':     "Optimize ('vacuum') database after processing.",
        'solo':         "Don't download crowd-sourced market data. (Implies '-O skipvend', supercedes '-O all', '-O clean', '-O listings'.)",
        '7days':        "Ignore data more than 7 days old during import, and expire old records after import.",
    }
    
    def __init__(self, tdb, tdenv):
        super().__init__(tdb, tdenv)
        
        self.dataPath = os.environ.get('TD_EDDB') or self.tdenv.tmpDir
        self.categoriesPath = Path("Category.csv")
        self.commoditiesPath = Path("Item.csv")
        self.rareItemPath = Path("RareItem.csv")
        self.shipPath = Path("Ship.csv")
        self.urlShipyard = "https://raw.githubusercontent.com/EDCD/FDevIDs/master/shipyard.csv"
        self.FDevShipyardPath = self.tdb.dataPath / Path("FDevShipyard.csv")
        self.shipVendorPath = Path("ShipVendor.csv")
        self.stationsPath = Path("Station.csv")
        self.sysPath = Path("System.csv")
        self.upgradesPath = Path("Upgrade.csv")
        self.urlOutfitting = "https://raw.githubusercontent.com/EDCD/FDevIDs/master/outfitting.csv"
        self.FDevOutfittingPath = self.tdb.dataPath / Path("FDevOutfitting.csv")
        self.upgradeVendorPath = Path("UpgradeVendor.csv")
        self.listingsPath = Path("listings.csv")
        self.liveListingsPath = Path("listings-live.csv")
        self.pricesPath = Path("listings.prices")
    
    def now(self):
        return datetime.datetime.now().strftime('%H:%M:%S')
    
    def downloadFile(self, path):
        """
        Fetch the latest dumpfile from the website if newer than local copy.
        """
        if path not in (self.liveListingsPath, self.listingsPath):
            localPath = Path(self.tdb.dataPath, path)
        else:
            localPath = Path(self.dataPath, path)
        
        url = BASE_URL + str(path)
        
        self.tdenv.NOTE("Checking for update to '{}'.", path)
        # Use an HTTP Request header to obtain the Last-Modified and Content-Length headers.
        # Also, tell the server to give us the un-compressed length of the file by saying
        # that >this< request only wants text.
        headers = {"User-Agent": "Trade-Dangerous", "Accept-Encoding": "identity"}
        try:
            response = requests.head(url, headers=headers, timeout=70)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.tdenv.WARN("Problem with download:\n    URL: {}\n    Error: {}", url, str(e))
            return False
        
        last_modified = response.headers.get("last-modified")
        dump_mod_time = parsedate_to_datetime(last_modified).timestamp()
        
        if Path.exists(localPath):
            local_mod_time = localPath.stat().st_mtime
            if local_mod_time >= dump_mod_time:
                self.tdenv.DEBUG0("'{}': Dump is not more recent than Local.", path)
                return False
        
        # The server doesn't know the gzip'd length, and we won't see the gzip'd data,
        # so we want the actual text-only length. Capture it here so we can tell the
        # transfer mechanism how big the file is going to be.
        length = response.headers.get("content-length")
        
        self.tdenv.NOTE("Downloading file '{}'.", path)
        transfers.download(self.tdenv, url, localPath, chunkSize=16384, length=length)
        
        # Change the timestamps on the file so they match the website
        os.utime(localPath, (dump_mod_time, dump_mod_time))
        
        return True
    
    def purgeSystems(self):
        """
        Purges systems from the System table that do not have any stations claiming to be in them.
        Keeps table from becoming too large because of fleet carriers moving to unpopulated systems.
        """
        self.tdenv.NOTE("Purging Systems with no stations: Start time = {}", self.now())
        
        Session = self.tdb.Session
        with Session.begin() as session:
            subq = select(SA.Station.system_id).where(SA.Station.system_id == SA.System.system_id)
            stmt = delete(SA.System).where(~exists(subq))
            session.execute(stmt)
        
        self.tdenv.NOTE("Finished purging Systems. End time = {}", self.now())
    
    def importListings(self, listings_file):
        """
        Updates the market data (StationItem) using `listings_file`.
        
        Rules:
          - If a row doesn't exist in DB → insert (copy CSV exactly).
          - If it exists → update only when CSV.modified > DB.modified.
          - If CSV.modified <= DB.modified → do nothing (no field changes).
        """
        listings_path = Path(self.dataPath, listings_file).absolute()
        from_live = listings_path != Path(self.dataPath, self.listingsPath).absolute()
        
        self.tdenv.NOTE("Checking listings")
        total = _count_listing_entries(self.tdenv, listings_path)
        if not total:
            self.tdenv.NOTE("No listings")
            return
        
        self.tdenv.NOTE(
            "Processing market data from {}: Start time = {}, Live = {}",
            listings_file, self.now(), from_live
        )
        
        Session = self.tdb.Session
        
        # Prefetch item/station IDs for early filtering
        with Session.begin() as session:
            item_lookup = _make_item_id_lookup(self.tdenv, session)
            station_lookup = _make_station_id_lookup(self.tdenv, session)
        
        is_debug = self.tdenv.debug > 0
        self.tdenv.DEBUG0("Processing entries...")
        
        with pbar.Progress(total, 40, prefix="Processing", style=pbar.LongRunningCountBar) as prog, \
               listings_path.open("r", encoding="utf-8", errors="ignore") as fh, \
               Session() as session:
            
            token = begin_bulk_mode(session, profile="eddblink", phase="incremental")
            try:
                commit_batch = get_import_batch_size(session, profile="eddblink")
                execute_batch = commit_batch or 10000  # cap statement size even if single final commit
                
                # Upsert: keys + guarded fields (including from_live), guarded by 'modified'
                table = SA.StationItem.__table__
                key_cols = ("station_id", "item_id")
                update_cols = (
                    "demand_price", "demand_units", "demand_level",
                    "supply_price", "supply_units", "supply_level",
                    "from_live",
                )
                upsert = get_upsert_fn(
                    session,
                    table,
                    key_cols=key_cols,
                    update_cols=update_cols,
                    modified_col="modified",
                    always_update=(),   # IMPORTANT: no unconditional updates
                )
                
                batch_rows = []
                since_commit = 0
                
                # optimize away millions of lookups
                def bump_progress():
                    prog.increment(1)
                
                from_timestamp = datetime.datetime.fromtimestamp
                utc = datetime.timezone.utc
                from_live_val = int(from_live)
                week_in_seconds = 7 * 24 * 60 * 60
                time_cutoff = 0 if not self.getOption("7days") else time.time() - week_in_seconds
                
                # Columns:
                #
                #   id, station_id, commodity_id, supply, supply_bracket, buy_price, sell_price, demand, demand_bracket, collected_at
                #   0   1           2             3       4               5          6           7       8               9
                reader = iter(csv.reader(fh))
                headers = next(reader)
                assert headers[:10] == ["id","station_id","commodity_id","supply","supply_bracket","buy_price","sell_price","demand","demand_bracket","collected_at"], "unrecognized listings csv format"
                
                for listing in reader:
                    bump_progress()
                    try:
                        station_id = int(listing[1])
                        if station_id not in station_lookup:
                            continue
                        
                        item_id = int(listing[2])
                        if item_id not in item_lookup:
                            continue  # skip rare items (not in Item table)
                        
                        listing_time = int(listing[9])
                        if listing_time < time_cutoff:
                            continue
                        dt_listing_time = from_timestamp(listing_time, utc)
                        
                        row = {
                            "station_id":   station_id,
                            "item_id":      item_id,
                            "modified":     dt_listing_time,   # guard column
                            "from_live":    from_live_val,     # copied exactly when updating/inserting
                            "supply_units": int(listing[3]),
                            "supply_level": int(listing[4]),
                            "supply_price": int(listing[5]),
                            "demand_price": int(listing[6]),
                            "demand_units": int(listing[7]),
                            "demand_level": int(listing[8]),
                        }
                        batch_rows += [row]
                        since_commit += 1
                        
                        if len(batch_rows) >= execute_batch:
                            upsert(batch_rows)
                            batch_rows[:] = []  # in-place clear without lookup
                        
                        if commit_batch and since_commit >= commit_batch:
                            session.commit()
                            since_commit = 0
                    
                    except Exception as e:  # pylint: disable=broad-exception-caught
                        self.tdenv.WARN("Bad listing row (skipped): {}  error: {}", listing, e)
                        continue
                
                if batch_rows:
                    upsert(batch_rows)
                    batch_rows[:] = []  # in-place clear
                
                session.commit()
            
            finally:
                end_bulk_mode(session, token)
        
        # with pbar.Progress(1, 40, prefix="Saving"):
        #     pass
        
        if self.getOption("7days"):
            with pbar.Progress(1, 40, prefix="Expiring") as prog, Session.begin() as session:
                prog.increment(1)
                session.execute(text("DELETE FROM StationItem WHERE modified < datetime('now', '-7 days')"))
        
        if self.getOption("optimize"):
            with pbar.Progress(1, 40, prefix="Optimizing"):
                if self.tdb.engine.dialect.name == "sqlite":
                    with Session.begin() as session:
                        prog.increment(1)
                        session.execute(text("VACUUM"))
        
        self.tdenv.NOTE("Finished processing market data. End time = {}", self.now())
    
    def run(self):
        """
        EDDN/EDDB link importer.
        
        Refactored DB flow:
          - No dialect-specific logic in the plugin.
          - Preflight uses TradeDB.reloadCache() (which centralizes sanity via lifecycle.ensure_fresh_db).
          - For '--clean' → do a single full rebuild with the RareItem dance.
          - Otherwise, if static CSVs changed → incrementally import only those tables (no drop/recreate).
          - Listings import and .prices regeneration unchanged.
        """
        self.tdenv.ignoreUnknown = True
        self.tdb.dataPath.mkdir(parents=True, exist_ok=True)
        
        # Enable 'listings' by default unless other explicit options are present
        default = True
        for option in self.options:
            if option not in ('force', 'skipvend', 'purge'):
                default = False
        if default:
            self.options["listings"] = True
        
        # Check if database already exists and enable `clean` if not.
        if lifecycle.is_empty(self.tdb.engine):
            self.options["clean"] = True
        
        if self.getOption("clean"):
            # Remove CSVs so downloads become the new source of truth
            for name in [
                "Category", "Item", "RareItem",
                "Ship", "ShipVendor",
                "Station", "System",
                "Upgrade", "UpgradeVendor",
                "FDevShipyard", "FDevOutfitting",
            ]:
                f = self.tdb.dataPath / f"{name}.csv"
                try:
                    os.remove(str(f))
                except FileNotFoundError:
                    pass
            
            # Remove .prices (DEPRECATED)
            try:
                os.remove(str(self.tdb.dataPath / "TradeDangerous.prices"))
            except FileNotFoundError:
                pass
            
            self.options["all"] = True
            self.options["force"] = True
        
        
        # Select which options will be updated
        if self.getOption("listings"):
            self.options["item"] = True
            self.options["station"] = True
        
        if self.getOption("shipvend"):
            self.options["ship"] = True
            self.options["station"] = True
        
        if self.getOption("upvend"):
            self.options["upgrade"] = True
            self.options["station"] = True
        
        if self.getOption("item"):
            self.options["station"] = True
        
        if self.getOption("rare"):
            self.options["station"] = True
        
        if self.getOption("station"):
            self.options["system"] = True
        
        if self.getOption("all"):
            self.options["item"] = True
            self.options["rare"] = True
            self.options["ship"] = True
            self.options["shipvend"] = True
            self.options["station"] = True
            self.options["system"] = True
            self.options["upgrade"] = True
            self.options["upvend"] = True
            self.options["listings"] = True
        
        if self.getOption("solo"):
            self.options["listings"] = False
            self.options["skipvend"] = True
        
        if self.getOption("skipvend"):
            self.options["shipvend"] = False
            self.options["upvend"] = False
        
        # Download required files and update tables.
        buildCache = False
        if self.getOption("upgrade"):
            if self.downloadFile(self.upgradesPath) or self.getOption("force"):
                transfers.download(self.tdenv, self.urlOutfitting, self.FDevOutfittingPath)
                buildCache = True
        
        if self.getOption("ship"):
            if self.downloadFile(self.shipPath) or self.getOption("force"):
                transfers.download(self.tdenv, self.urlShipyard, self.FDevShipyardPath)
                buildCache = True
        
        if self.getOption("rare"):
            if self.downloadFile(self.rareItemPath) or self.getOption("force"):
                buildCache = True
        
        if self.getOption("shipvend"):
            if self.downloadFile(self.shipVendorPath) or self.getOption("force"):
                buildCache = True
        
        if self.getOption("upvend"):
            if self.downloadFile(self.upgradeVendorPath) or self.getOption("force"):
                buildCache = True
        
        if self.getOption("system"):
            if self.downloadFile(self.sysPath) or self.getOption("force"):
                buildCache = True
        
        if self.getOption("station"):
            if self.downloadFile(self.stationsPath) or self.getOption("force"):
                buildCache = True
        
        if self.getOption("item"):
            if self.downloadFile(self.commoditiesPath) or self.getOption("force"):
                self.downloadFile(self.categoriesPath)
                buildCache = True
        
        modified = buildCache
        
        # Remake the .db files with the updated info.
        if buildCache:
            self.tdb.close()
            self.tdb.reloadCache()
            if self.tdb.engine.dialect.name == "sqlite":
                # kfsone: see https://sqlite.org/pragma.html#pragma_optimize
                self.tdb.Session().execute(text("PRAGMA optimize=0x10002"))
            self.tdb.close()
        
        if self.getOption("purge"):
            self.purgeSystems()
            modified = True
        
        # Listings import (prices)
        if self.getOption("listings"):
            if self.downloadFile(self.listingsPath) or self.getOption("force"):
                self.importListings(self.listingsPath)
                modified = True
            if self.downloadFile(self.liveListingsPath) or self.getOption("force"):
                self.importListings(self.liveListingsPath)
                modified = True
        
        # if self.getOption("listings"):
        #     self.tdenv.NOTE("Regenerating .prices file.")
        #     cache.regeneratePricesFile(self.tdb, self.tdenv)

        if self.tdb.engine.dialect.name == "sqlite":
            with self.tdb.Session.begin() as session:
                if self.getOption("optimize"):
                    with bench("Vacuum and optimize", self.tdenv):
                        session.execute(text("VACUUM"))
                        # This is a very aggressive analyze/optimize pass
                        session.execute(text("ANALYZE"))
                else:
                    with bench("DB Tuning", self.tdenv):
                        session.execute(text("PRAGMA optimize"))
                    self.tdenv.INFO("Use --opt=optimize periodically for better query performance")
    
        self.tdenv.NOTE("Import completed.")
        
        if modified:
            self.tdb.removePersist()
        
        return False



from __future__ import annotations
import configparser
from pathlib import Path
from typing import Any, Dict

DEFAULTS: Dict[str, Dict[str, Any]] = {
    "database": {"backend": "sqlite"},
    "mariadb": {
        "host": "127.0.0.1",
        "port": 3306,
        "user": "",
        "password": "",
        "name": "tradedangerous",
        "driver": "mariadbconnector",  # or 'pymysql'
        "charset": "utf8mb4",
    },
    "sqlite": {"sqlite_filename": "TradeDangerous.db"},
    "paths": {"data_dir": "./data", "tmp_dir": "./tmp"},
    "engine": {
        "pool_size": 10,
        "max_overflow": 20,
        "pool_timeout": 30,
        "pool_recycle": 1800,
        "isolation_level": "READ COMMITTED",
        "echo": False,
        "connect_timeout": 10,
    },
}
# --- Runtime default path correction ----------------------------------------
# Convert relative defaults ("./data", "./tmp") into absolute paths under the
# current working directory. This prevents first-run installs from resolving
# relative to the package install directory or venv when no db_config.ini exists.
try:
    _cwd = Path.cwd()
    DEFAULTS["paths"]["data_dir"] = str((_cwd / "data").resolve())
    DEFAULTS["paths"]["tmp_dir"]  = str((_cwd / "tmp").resolve())
except Exception:
    # Best effort; fall back to shipped defaults if CWD is inaccessible
    pass
# ---------------------------------------------------------------------------

# Hardened parser: allow inline comments and disable interpolation
CFG_KW = dict(inline_comment_prefixes=(";", "#"), interpolation=None)

def _parse_bool(s: str) -> bool:
    return str(s).strip().lower() in {"1", "true", "yes", "on"}

def _as_int(s: str, default: int | None = None) -> int | None:
    try:
        return int(str(s).strip())
    except (TypeError, ValueError):
        return default

def _coerce_types(d: Dict[str, Any]) -> Dict[str, Any]:
    eng = d.get("engine", {})
    if "echo" in eng:
        eng["echo"] = _parse_bool(eng["echo"]) if isinstance(eng["echo"], str) else bool(eng["echo"])
    for k in ("pool_size", "max_overflow", "pool_timeout", "pool_recycle", "connect_timeout"):
        if k in eng:
            eng[k] = _as_int(eng[k], DEFAULTS["engine"][k])
    if "mariadb" in d and "port" in d["mariadb"]:
        d["mariadb"]["port"] = _as_int(d["mariadb"]["port"], DEFAULTS["mariadb"]["port"])
    return d

def load_config(path: str | Path | None = None) -> Dict[str, Any]:
    """Load configuration as a dict with typed values.
    Search order:
      1) explicit *path* if provided
      2) TD_DB_CONFIG env (if file exists)
      3) ./db_config.ini (cwd)
      4) in-code DEFAULTS
    """
    cfg_path: Path | None = None
    if path is not None:
        p = Path(path)
        if p.exists():
            cfg_path = p
    else:
        # Prefer environment variable if it points to an existing file
        try:
            from .paths import resolve_db_config_path
            env_candidate = resolve_db_config_path()
            if env_candidate.exists():
                cfg_path = env_candidate
        except Exception:
            # If anything goes wrong resolving the env, fall back to defaults below
            pass
        
        # Fall back to local file in CWD
        if cfg_path is None:
            p = Path.cwd() / "db_config.ini"
            if p.exists():
                cfg_path = p
    
    # start with defaults
    result: Dict[str, Any] = {k: (v.copy() if isinstance(v, dict) else v) for k, v in DEFAULTS.items()}
    
    if cfg_path:
        parser = configparser.ConfigParser(**CFG_KW)
        with cfg_path.open("r", encoding="utf-8") as fh:
            parser.read_file(fh)
        for section in parser.sections():
            result.setdefault(section, {})
            for key, val in parser.items(section):
                result[section][key] = val
    
    return _coerce_types(result)



# tradedangerous/commands/shipvendor_cmd.py — DEPRECATED (no-op)
# This command no longer maintains ship lists directly.
# Guidance:
#   • Import authoritative data via:   trade import -P spansh  |  trade import -P eddblink
#   • Search for ships via:            trade buy --near "<place>" --ly N "<ship>"

from __future__ import annotations
from .parsing import ParseArgument  # to swallow arbitrary positional args

# ---- Command metadata ----
help = "DEPRECATED: no longer used. See deprecation banner when run."
name = "shipvendor"
epilog = None
acceptUnknown = True

# No DB access needed.
wantsTradeDB = False
usesTradeData = False

# Accept ANY number of positional args and ignore them (prevents parser errors).
arguments = (
    ParseArgument(
        'args',
        help="(deprecated) ignored",
        nargs='*',
        type=str,
    ),
)
# No switches; unknown switches will still be rejected by the global parser.
switches = tuple()


def _banner() -> str:
    return (
        "\n"
        "=== DEPRECATION NOTICE: shipvendor ==================================\n"
        "This command is no longer used and does not modify the database.\n"
        "• Import authoritative data with:  trade import -P eddblink | -P spansh\n"
        "• Search for ships using:         trade buy --near \"<place>\" --ly N \"<ship>\"\n"
        "======================================================================\n"
    )


def run(results, cmdenv, tdb=None):
    """
    No-op implementation: print banner and exit immediately.
    All arguments/switches are ignored by design.
    """
    banner = _banner()
    try:
        cmdenv.NOTE("{}", banner)
    except Exception:
        print(banner)
    return None


def render(results, cmdenv, tdb=None):
    # No output beyond the banner emitted in run().
    return



"""This module should handle filesystem related operations
"""
from os import makedirs, PathLike
from pathlib import Path
from shutil import copy as shcopy
import typing

__all__ = ['copy', 'copyallfiles', 'touch', 'ensurefolder', 'file_line_count']

Pathlike: typing.TypeAlias = Path | str


def pathify(*args: str | Path) -> Path:
    """ pathify will ensure a Path given input(s) that can be
        either a single Path/str, or components to form one.
        e.g.
            pathify("C:/temp") -> Path("c:/temp")
            pathify("C":", "temp") -> Path("C:/temp")
            pathify("C:", Path("temp")) -> Path("C:/temp")
    """
    if len(args) > 1 or not isinstance(args[0], Path):
        return Path(*args)
    return args[0]

def copy(src: PathLike, dst: PathLike) -> Path:
    """
    copy src to dst
    takes string or Path object as input
    returns Path(dst) on success
    raises FileNotFoundError if src does not exist
    """
    srcPath = pathify(src).resolve()
    dstPath = pathify(dst)
    shcopy(str(srcPath), str(dstPath))
    return dstPath


def copy_if_newer(src: Pathlike, dst: Pathlike) -> Path:
    """
    copy src to dst if src is newer
    takes string or Path object as input
    returns Path(dst) on success
    returns Path(src) if not newer
    raises FileNotFoundError if src does not exist
    """
    srcPath = pathify(src).resolve()
    dstPath = pathify(dst)
    if dstPath.exists() and dstPath.stat().st_mtime >= srcPath.stat().st_mtime:
        return srcPath
    
    shcopy(srcPath, dstPath)    # kfs: python 3.10 don't need us to strify these
    return dstPath

def copyallfiles(srcdir: Pathlike, dstdir: Pathlike) -> None:
    """
    Copies all files in srcdir to dstdir
    """
    srcPath = pathify(srcdir)
    dstPath = pathify(dstdir)
    
    for p in srcPath.glob('*.*'):
        if p.is_file():
            copy(p, dstPath / p.name)

def touch(filename: Pathlike) -> Path:
    """
    Creates file if it doesn't exist.
    Always modifies utime.
    Returns a Path(filename)
    """
    path = pathify(filename)
    path.touch(exist_ok=True)
    return path

def ensurefolder(folder: Pathlike) -> Path:
    """Creates the folder if it doesn't exist
    
    Parameters
    ----------
    folder : string|pathlib.Path
        path to the folder
    
    Returns
    -------
    pathlib.Path
        Resolved path of the folder
    """
    folderPath = pathify(folder)
    try:
        makedirs(str(folderPath))
    except FileExistsError:
        pass
    return folderPath.resolve()


def file_line_count(from_file: Pathlike, buf_size: int = 128 * 1024, *, missing_ok: bool = False) -> int:
    """ counts the number of newline characters in a given file. """
    if not isinstance(from_file, Path):
        from_file = Path(from_file)
    
    if missing_ok and not from_file.exists():
        return 0
    
    # Pre-allocate a buffer so that we aren't putting pressure on the garbage collector.
    buf = bytearray(buf_size)
    
    # Capture it's counting method, so we don't have to keep looking that up on
    # large files.
    counter = buf.count
    
    total = 0
    with from_file.open("rb") as fh:
        # Capture the 'readinto' method to avoid lookups.
        reader = fh.readinto
        
        # read into the buffer and capture the number of bytes fetched,
        # which will be 'size' until the last read from the file.
        read = reader(buf)
        while read == buf_size:  # nominal case for large files
            total += counter(b'\n')
            read = reader(buf)
        
        # when 0 <= read < buf_size we're on the last page of the
        # file, so we need to take a slice of the buffer, which creates
        # a new object, thus we also have to lookup count. it's trivial
        # but if you have to do it 10,000x it's definitely not a rounding error.
        return total + buf[:read].count(b'\n')





# tradedangerous/db/orm_models.py
from __future__ import annotations

from typing import Optional

from sqlalchemy import (
    MetaData,
    ForeignKey,
    Integer,
    BigInteger,
    String,
    CHAR,
    Enum,
    Index,
    UniqueConstraint,
    CheckConstraint,
    text,
    Column,
    DateTime,
)
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship, object_session
from sqlalchemy.sql import expression
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.types import TypeDecorator


# ---- Dialect-aware time utilities (moved before model usage) ----
class now6(expression.FunctionElement):
    """CURRENT_TIMESTAMP with microseconds on MySQL/MariaDB; plain CURRENT_TIMESTAMP elsewhere."""
    type = DateTime()
    inherit_cache = True


@compiles(now6, "mysql")
@compiles(now6, "mariadb")
def _mysql_now6(element, compiler, **kw):
    return "CURRENT_TIMESTAMP(6)"


@compiles(now6)
def _default_now(element, compiler, **kw):
    return "CURRENT_TIMESTAMP"


class DateTime6(TypeDecorator):
    """DATETIME that is DATETIME(6) on MySQL/MariaDB, generic DateTime elsewhere."""
    impl = DateTime
    cache_ok = True
    
    def load_dialect_impl(self, dialect):
        if dialect.name in ("mysql", "mariadb"):
            from sqlalchemy.dialects.mysql import DATETIME as _MYSQL_DATETIME
            return dialect.type_descriptor(_MYSQL_DATETIME(fsp=6))
        return dialect.type_descriptor(DateTime())


# ---------- Dialect Helpers --------
class CIString(TypeDecorator):
    """
    Case-insensitive string type.
    - SQLite → uses NOCASE collation
    - MySQL/MariaDB → uses utf8mb4_unicode_ci
    - Others → plain String
    """
    impl = String
    cache_ok = True
    
    def __init__(self, length, **kwargs):
        super().__init__(length=length, **kwargs)
    
    def load_dialect_impl(self, dialect):
        if dialect.name == "sqlite":
            return dialect.type_descriptor(String(self.impl.length, collation="NOCASE"))
        elif dialect.name in ("mysql", "mariadb"):
            return dialect.type_descriptor(String(self.impl.length, collation="utf8mb4_unicode_ci"))
        else:
            return dialect.type_descriptor(String(self.impl.length))


# ---------- Naming & Base ----------
naming_convention = {
    "ix": "ix_%(table_name)s__%(column_0_N_name)s",
    "uq": "uq_%(table_name)s__%(column_0_N_name)s",
    "ck": "ck_%(table_name)s__%(column_0_name)s",   # use column name, not constraint_name
    "fk": "fk_%(table_name)s__%(column_0_N_name)s__%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}
metadata = MetaData(naming_convention=naming_convention)


class Base(DeclarativeBase):
    metadata = metadata

# ---------- Enums ----------
TriState = Enum(
    "Y",
    "N",
    "?",
    native_enum=False,
    create_constraint=True,
    validate_strings=True,
)

PadSize = Enum(
    "S",
    "M",
    "L",
    "?",
    native_enum=False,
    create_constraint=True,
    validate_strings=True,
)


# ---------- Core Domain ----------
class Added(Base):
    __tablename__ = "Added"
    
    added_id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    name: Mapped[str] = mapped_column(CIString(128), nullable=False, unique=True)
    
    # Relationships
    systems: Mapped[list["System"]] = relationship(back_populates="added")


class System(Base):
    __tablename__ = "System"
    
    system_id: Mapped[int] = mapped_column(BigInteger, primary_key=True)
    name: Mapped[str] = mapped_column(CIString(128), nullable=False)
    pos_x: Mapped[float] = mapped_column(nullable=False)
    pos_y: Mapped[float] = mapped_column(nullable=False)
    pos_z: Mapped[float] = mapped_column(nullable=False)
    added_id: Mapped[int | None] = mapped_column(
        ForeignKey("Added.added_id", onupdate="CASCADE", ondelete="CASCADE")
    )
    modified: Mapped[str] = mapped_column(
        DateTime6(),
        server_default=now6(),
        onupdate=now6(),
        nullable=False,
    )
    
    # Relationships
    added: Mapped[Optional["Added"]] = relationship(back_populates="systems")
    stations: Mapped[list["Station"]] = relationship(back_populates="system", cascade="all, delete-orphan")
    
    __table_args__ = (
        Index("idx_system_by_pos", "pos_x", "pos_y", "pos_z", "system_id"),
        Index("idx_system_by_name", "name"),
    )




class Station(Base):
    __tablename__ = "Station"
    
    station_id: Mapped[int] = mapped_column(BigInteger, primary_key=True)
    name: Mapped[str] = mapped_column(CIString(128), nullable=False)
    
    # type widened; cascade semantics unchanged (DELETE only)
    system_id: Mapped[int] = mapped_column(
        BigInteger,
        ForeignKey("System.system_id", ondelete="CASCADE"),
        nullable=False,
    )
    
    ls_from_star: Mapped[int] = mapped_column(Integer, nullable=False, server_default=text("0"))
    
    blackmarket: Mapped[str] = mapped_column(TriState, nullable=False, server_default=text("'?'"))
    max_pad_size: Mapped[str] = mapped_column(PadSize, nullable=False, server_default=text("'?'"))
    market: Mapped[str] = mapped_column(TriState, nullable=False, server_default=text("'?'"))
    shipyard: Mapped[str] = mapped_column(TriState, nullable=False, server_default=text("'?'"))
    outfitting: Mapped[str] = mapped_column(TriState, nullable=False, server_default=text("'?'"))
    rearm: Mapped[str] = mapped_column(TriState, nullable=False, server_default=text("'?'"))
    refuel: Mapped[str] = mapped_column(TriState, nullable=False, server_default=text("'?'"))
    repair: Mapped[str] = mapped_column(TriState, nullable=False, server_default=text("'?'"))
    planetary: Mapped[str] = mapped_column(TriState, nullable=False, server_default=text("'?'"))
    
    type_id: Mapped[int] = mapped_column(Integer, nullable=False, server_default=text("0"))
    modified: Mapped[str] = mapped_column(DateTime6(), server_default=now6(), onupdate=now6(), nullable=False)
    
    # Relationships
    system: Mapped["System"] = relationship(back_populates="stations")
    items: Mapped[list["StationItem"]] = relationship(back_populates="station", cascade="all, delete-orphan")
    ship_vendors: Mapped[list["ShipVendor"]] = relationship(back_populates="station", cascade="all, delete-orphan")
    upgrade_vendors: Mapped[list["UpgradeVendor"]] = relationship(back_populates="station", cascade="all, delete-orphan")
    
    __table_args__ = (
        Index("idx_station_by_system", "system_id"),
        Index("idx_station_by_name", "name"),
    )




class Category(Base):
    __tablename__ = "Category"
    
    category_id: Mapped[int] = mapped_column(Integer, primary_key=True)
    name: Mapped[str] = mapped_column(CIString(128), nullable=False)
    
    # Relationships
    items: Mapped[list["Item"]] = relationship(back_populates="category")
    
    __table_args__ = (Index("idx_category_by_name", "name"),)


class Item(Base):
    __tablename__ = "Item"
    
    item_id: Mapped[int] = mapped_column(Integer, primary_key=True)
    name: Mapped[str] = mapped_column(CIString(128), nullable=False)
    category_id: Mapped[int] = mapped_column(
        ForeignKey("Category.category_id", onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
    )
    ui_order: Mapped[int] = mapped_column(Integer, nullable=False, server_default=text("0"))
    avg_price: Mapped[int | None] = mapped_column(Integer)
    fdev_id: Mapped[int | None] = mapped_column(Integer)
    
    # Relationships
    category: Mapped["Category"] = relationship(back_populates="items")
    stations: Mapped[list["StationItem"]] = relationship(back_populates="item", cascade="all, delete-orphan")
    
    __table_args__ = (
        Index("idx_item_by_fdevid", "fdev_id"),
        Index("idx_item_by_category", "category_id"),
    )



class StationItem(Base):
    __tablename__ = "StationItem"
    
    station_id: Mapped[int] = mapped_column(
        BigInteger,
        ForeignKey("Station.station_id", ondelete="CASCADE", onupdate="CASCADE"),
        primary_key=True,
    )
    item_id: Mapped[int] = mapped_column(
        ForeignKey("Item.item_id", ondelete="CASCADE", onupdate="CASCADE"),
        primary_key=True,
    )
    demand_price: Mapped[int] = mapped_column(Integer, nullable=False)
    demand_units: Mapped[int] = mapped_column(Integer, nullable=False)
    demand_level: Mapped[int] = mapped_column(Integer, nullable=False)
    supply_price: Mapped[int] = mapped_column(Integer, nullable=False)
    supply_units: Mapped[int] = mapped_column(Integer, nullable=False)
    supply_level: Mapped[int] = mapped_column(Integer, nullable=False)
    modified: Mapped[str] = mapped_column(DateTime6(), server_default=now6(), onupdate=now6(), nullable=False)
    from_live: Mapped[int] = mapped_column(Integer, nullable=False, server_default=text("0"))
    
    # Relationships
    station: Mapped["Station"] = relationship(back_populates="items")
    item: Mapped["Item"] = relationship(back_populates="stations")
    
    __table_args__ = (
        Index("si_mod_stn_itm", "modified", "station_id", "item_id"),
        Index("si_itm_dmdpr", "item_id", "demand_price", sqlite_where=text("demand_price > 0")),
        Index("si_itm_suppr", "item_id", "supply_price", sqlite_where=text("supply_price > 0")),
        {"sqlite_with_rowid": False},
    )




class Ship(Base):
    __tablename__ = "Ship"
    
    ship_id: Mapped[int] = mapped_column(Integer, primary_key=True)
    name: Mapped[str] = mapped_column(CIString(128), nullable=False)
    cost: Mapped[int | None] = mapped_column(Integer)
    
    # Relationships
    vendors: Mapped[list["ShipVendor"]] = relationship(back_populates="ship")


class ShipVendor(Base):
    __tablename__ = "ShipVendor"
    
    ship_id: Mapped[int] = mapped_column(
        ForeignKey("Ship.ship_id", ondelete="CASCADE", onupdate="CASCADE"),
        primary_key=True,
    )
    station_id: Mapped[int] = mapped_column(
        BigInteger,
        ForeignKey("Station.station_id", ondelete="CASCADE", onupdate="CASCADE"),
        primary_key=True,
    )
    modified: Mapped[str] = mapped_column(DateTime6(), server_default=now6(), onupdate=now6(), nullable=False)
    
    # Relationships
    ship: Mapped["Ship"] = relationship(back_populates="vendors")
    station: Mapped["Station"] = relationship(back_populates="ship_vendors")
    
    __table_args__ = (Index("idx_shipvendor_by_station", "station_id"),)


class Upgrade(Base):
    __tablename__ = "Upgrade"
    
    upgrade_id: Mapped[int] = mapped_column(Integer, primary_key=True)
    name: Mapped[str] = mapped_column(CIString(128), nullable=False)
    class_: Mapped[int] = mapped_column("class", Integer, nullable=False)
    rating: Mapped[str] = mapped_column(CHAR(1), nullable=False)
    ship: Mapped[str | None] = mapped_column(CIString(128))
    
    # Relationships
    vendors: Mapped[list["UpgradeVendor"]] = relationship(back_populates="upgrade")


class UpgradeVendor(Base):
    __tablename__ = "UpgradeVendor"
    
    upgrade_id: Mapped[int] = mapped_column(
        ForeignKey("Upgrade.upgrade_id", ondelete="CASCADE", onupdate="CASCADE"),
        primary_key=True,
    )
    station_id: Mapped[int] = mapped_column(
        BigInteger,
        ForeignKey("Station.station_id", ondelete="CASCADE", onupdate="CASCADE"),
        primary_key=True,
    )
    modified: Mapped[str] = mapped_column(DateTime6(), nullable=False, server_default=now6(), onupdate=now6())
    
    # Relationships
    upgrade: Mapped["Upgrade"] = relationship(back_populates="vendors")
    station: Mapped["Station"] = relationship(back_populates="upgrade_vendors")
    
    __table_args__ = (Index("idx_vendor_by_station_id", "station_id"),)


class RareItem(Base):
    __tablename__ = "RareItem"
    
    rare_id: Mapped[int] = mapped_column(Integer, primary_key=True)
    station_id: Mapped[int] = mapped_column(
        BigInteger,
        ForeignKey("Station.station_id", ondelete="CASCADE", onupdate="CASCADE"),
        nullable=False,
    )
    category_id: Mapped[int] = mapped_column(
        ForeignKey("Category.category_id", onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
    )
    name: Mapped[str] = mapped_column(CIString(128), nullable=False)
    cost: Mapped[int | None] = mapped_column(Integer)
    max_allocation: Mapped[int | None] = mapped_column(Integer)
    illegal: Mapped[str] = mapped_column(TriState, nullable=False, server_default=text("'?'"))
    suppressed: Mapped[str] = mapped_column(TriState, nullable=False, server_default=text("'?'"))
    
    __table_args__ = (UniqueConstraint("name", name="uq_rareitem_name"),)




class FDevShipyard(Base):
    __tablename__ = "FDevShipyard"
    
    id = Column(Integer, primary_key=True, unique=True, nullable=False)
    symbol = Column(CIString(128))
    name = Column(CIString(128))
    entitlement = Column(String(50))


class FDevOutfitting(Base):
    __tablename__ = "FDevOutfitting"
    
    id = Column(Integer, primary_key=True, unique=True, nullable=False)
    symbol = Column(CIString(128))
    category = Column(String(10))
    name = Column(CIString(128))
    mount = Column(String(20))
    guidance = Column(String(20))
    ship = Column(CIString(128))
    class_ = Column("class", String(1), nullable=False)
    rating = Column(String(1), nullable=False)
    entitlement = Column(String(50))
    
    __table_args__ = (
        CheckConstraint(
            "category IN ('hardpoint','internal','standard','utility')",
            name="ck_fdo_category",
        ),
        CheckConstraint(
            "(mount IN ('Fixed','Gimballed','Turreted')) OR (mount IS NULL)",
            name="ck_fdo_mount",
        ),
        CheckConstraint(
            "(guidance IN ('Dumbfire','Seeker','Swarm')) OR (guidance IS NULL)",
            name="ck_fdo_guidance",
        ),
    )


# ---------- Control & Staging ----------
class ExportControl(Base):
    """
    Singleton control row for hybrid export/watermarking.
    - id: always 1
    - last_full_dump_time: watermark
    - last_reset_key: optional cursor for chunked from_live resets
    """
    __tablename__ = "ExportControl"
    
    id: Mapped[int] = mapped_column(Integer, primary_key=True, server_default=text("1"))
    last_full_dump_time: Mapped[str] = mapped_column(DateTime6(), nullable=False)
    last_reset_key: Mapped[int | None] = mapped_column(BigInteger, nullable=True)


class StationItemStaging(Base):
    """
    Staging table for bulk loads (no FKs). Same columns as StationItem.
    """
    __tablename__ = "StationItem_staging"
    
    station_id: Mapped[int] = mapped_column(BigInteger, primary_key=True)
    item_id: Mapped[int] = mapped_column(Integer, primary_key=True)
    demand_price: Mapped[int] = mapped_column(Integer, nullable=False)
    demand_units: Mapped[int] = mapped_column(Integer, nullable=False)
    demand_level: Mapped[int] = mapped_column(Integer, nullable=False)
    supply_price: Mapped[int] = mapped_column(Integer, nullable=False)
    supply_units: Mapped[int] = mapped_column(Integer, nullable=False)
    supply_level: Mapped[int] = mapped_column(Integer, nullable=False)
    modified: Mapped[str] = mapped_column(DateTime6(), server_default=now6(), onupdate=now6(), nullable=False)
    from_live: Mapped[int] = mapped_column(Integer, nullable=False, server_default=text("0"))
    
    __table_args__ = (Index("idx_sistaging_stn_itm", "station_id", "item_id"),)




__all__ = [
    # Base
    "Base",
    # Core
    "Added",
    "System",
    "Station",
    "Category",
    "Item",
    "StationItem",
    "Ship",
    "ShipVendor",
    "Upgrade",
    "UpgradeVendor",
    "RareItem",
    "FDevShipyard",
    "FDevOutfitting",
    # Control & staging
    "ExportControl",
    "StationItemStaging",
]



from __future__ import annotations
from collections import defaultdict
from .commandenv import ResultRow
from .exceptions import CommandLineError, NoDataError
from ..formatting import RowFormat, max_len
from ..tradedb import Station, System, TradeDB
from .parsing import (
    AvoidPlacesArgument, BlackMarketSwitch, FleetCarrierArgument, MutuallyExclusiveGroup,
    NoPlanetSwitch, OdysseyArgument, PadSizeArgument, ParseArgument, PlanetaryArgument,
)
from sqlalchemy import text

# TODO: Add UPGRADE_MODE
ITEM_MODE = "Item"
SHIP_MODE = "Ship"

######################################################################
# Parser config

help = 'Find places to buy a given item within range of a given station.'
name = 'buy'
epilog = None
wantsTradeDB = True
arguments = (
    ParseArgument(
        'name',
        help = 'Items or Ships to look for.',
        type = str,
        nargs = '+',
    ),
)
switches = (
    ParseArgument(
        '--supply', '--quantity',
        help = 'Limit to stations known to have at least this much supply.',
        default = 0,
        type = int,
    ),
    ParseArgument(
        '--near',
        help = 'Find sellers within jump range of this system.',
        type = str
    ),
    ParseArgument(
        '--ly',
        help = '[Requires --near] Systems within this range of --near.',
        default = None,
        dest = 'maxLyPer',
        metavar = 'N.NN',
        type = float,
    ),
    ParseArgument(
        '--limit',
        help = 'Maximum number of results to list.',
        default = None,
        type = int,
    ),
    AvoidPlacesArgument(),
    ParseArgument('--age', '--max-days-old', '-MD',
        help = 'Maximum age (in days) of trade data to use.',
        metavar = 'DAYS',
        type = float,
        dest = 'maxAge',
    ),
    PadSizeArgument(),
    MutuallyExclusiveGroup(
        NoPlanetSwitch(),
        PlanetaryArgument(),
    ),
    FleetCarrierArgument(),
    OdysseyArgument(),
    BlackMarketSwitch(),
    MutuallyExclusiveGroup(
        ParseArgument(
            '--one-stop', '-1',
            help = 'Only list stations that carry all items listed.',
            action = 'store_true',
            dest = 'oneStop',
        ),
        ParseArgument(
            '--price-sort', '-P',
            help = '(When using --near) Sort by price not distance',
            action = 'store_true',
            default = False,
            dest = 'sortByPrice',
        ),
        ParseArgument(
            '--units-sort', '-S',
            help = 'Sort by available units followed by price',
            action = 'store_true',
            default = False,
            dest = 'sortByUnits',
        ),
    ),
    ParseArgument(
        '--gt',
        help = 'Limit to prices above Ncr',
        metavar = 'N',
        dest = 'gt',
        type = "credits",
    ),
    ParseArgument(
        '--lt',
        help = 'Limit to prices below Ncr',
        metavar = 'N',
        dest = 'lt',
        type = "credits",
    ),
    ParseArgument('--ls-max',
        help='Only consider stations up to this many ls from their star.',
        metavar='LS',
        dest='maxLs',
        type=int,
        default=0,
    ),
)


def get_lookup_list(cmdenv, tdb):
    # Credit: http://stackoverflow.com/a/952952/257645
    # Turns [['a'],['b','c']] => ['a', 'b', 'c']
    names = [
        name for names in cmdenv.name for name in names.split(',')
    ]
    # We only support searching for one type of purchase a time: ship or item.
    # Our first match is open-ended, but once we have matched one type of
    # thing, the remaining arguments are all sourced from the same pool.
    # Thus: [food, cobra, metals] is illegal but [metals, hydrogen] is legal.
    mode = None
    
    queries = {}
    for name in names:
        if mode is not SHIP_MODE:
            # Either no mode selected yet or we are in ITEM_MODE.
            # Consider categories first.
            try:
                category = tdb.lookupCategory(name)
                for item in category.items:
                    names.append(item.name())
                    queries[item.ID] = item
                mode = ITEM_MODE
                continue
            except LookupError:
                pass
            
            # Item names secondary.
            try:
                item = tdb.lookupItem(name)
                cmdenv.DEBUG0("Looking up item {} (#{})", item.name(), item.ID)
                queries[item.ID] = item
                mode = ITEM_MODE
                continue
            except LookupError:
                if mode is ITEM_MODE:
                    raise CommandLineError(
                        "Unrecognized item: {}".format(name)
                    )
                pass
        
        # Either no mode selected yet or we are in SHIP_MODE.
        try:
            ship = tdb.lookupShip(name)
            cmdenv.DEBUG0("Looking up ship {} (#{})", ship.name(), ship.ID)
            queries[ship.ID] = ship
            mode = SHIP_MODE
            continue
        except LookupError:
            if not mode:
                raise CommandLineError(
                    "Unrecognized item/ship: {}".format(name)
                )
            raise CommandLineError(
                "Unrecognized ship: {}".format(name)
            )
    
    return queries, mode


def sql_query(cmdenv, tdb, queries, mode):
    """
    Backend-portable query builder.
    - Uses named binds (':param') instead of SQLite '?'.
    - Materializes rows eagerly to avoid closed-cursor issues.
    - Preserves return shapes:
        * Ship:   (ship_id, station_id, cost, 1)
        * Item:   (item_id, station_id, supply_price, supply_units)
    """
    ids = list(queries.keys())
    
    # Build a stable, named-parameter IN(...) list
    params = {}
    placeholders = []
    for i, val in enumerate(ids):
        key = f"id{i}"
        placeholders.append(f":{key}")
        params[key] = val
    id_list_sql = ",".join(placeholders)
    
    if mode is SHIP_MODE:
        columns = "s.ship_id, s.station_id, sh.cost, 1"
        tables = "ShipVendor AS s JOIN Ship AS sh ON sh.ship_id = s.ship_id"
        constraints = [f"(s.ship_id IN ({id_list_sql}))"]
    else:
        columns = "s.item_id, s.station_id, s.supply_price, s.supply_units"
        tables = "StationItem AS s"
        constraints = [
            f"(s.item_id IN ({id_list_sql}))",
            "(s.supply_price > 0)",  # preserves index intent across backends
        ]
        if cmdenv.supply:
            constraints.append("(s.supply_units >= :supply)")
            params["supply"] = cmdenv.supply
        if cmdenv.lt:
            constraints.append("(s.supply_price < :lt)")
            params["lt"] = cmdenv.lt
        if cmdenv.gt:
            constraints.append("(s.supply_price > :gt)")
            params["gt"] = cmdenv.gt
    
    where_clause = " AND ".join(constraints)
    stmt = f"SELECT DISTINCT {columns} FROM {tables} WHERE {where_clause}"
    cmdenv.DEBUG0('SQL: {} ; params={}', stmt, params)
    
    # Eagerly fetch to avoid closed cursor when iterating later.
    with tdb.engine.connect() as conn:
        result = conn.execute(text(stmt), params)
        rows = result.fetchall()
    return rows



######################################################################
# Perform query and populate result set


def run(results, cmdenv, tdb):
    if cmdenv.lt and cmdenv.gt:
        if cmdenv.lt <= cmdenv.gt:
            raise CommandLineError("--gt must be lower than --lt")
    
    # Find out what we're looking for.
    queries, mode = get_lookup_list(cmdenv, tdb)
    cmdenv.DEBUG0("{} query: {}", mode, queries.values())
    
    avoidSystems = {s for s in cmdenv.avoidPlaces if isinstance(s, System)}
    avoidStations = {s for s in cmdenv.avoidPlaces if isinstance(s, Station)}
    
    # Summarize
    results.summary = ResultRow()
    results.summary.mode = mode
    results.summary.queries = queries
    results.summary.oneStop = cmdenv.oneStop
    results.summary.avoidSystems = avoidSystems
    results.summary.avoidStations = avoidStations
    
    # In single mode with detail enabled, add average reports.
    # Thus if you're looking up "algae" or the "asp", it'll
    # tell you the average/ship cost.
    singleMode = len(queries) == 1
    if singleMode and cmdenv.detail:
        first = list(queries.values())[0]
        if mode is SHIP_MODE:
            results.summary.avg = first.cost
        else:
            # Portable AVG with named bind; eager scalar fetch
            with tdb.engine.connect() as conn:
                avg_val = conn.execute(
                    text("""
                        SELECT AVG(si.supply_price) AS avg_price
                          FROM StationItem AS si
                         WHERE si.item_id = :item_id AND si.supply_price > 0
                    """),
                    {"item_id": first.ID},
                ).scalar()
            results.summary.avg = int(avg_val or 0)
    
    # System-based search
    nearSystem = cmdenv.nearSystem
    if nearSystem:
        maxLy = cmdenv.maxLyPer or tdb.maxSystemLinkLy
        results.summary.near = nearSystem
        results.summary.ly = maxLy
        distanceFn = nearSystem.distanceTo
    else:
        distanceFn = None
    
    oneStopMode = cmdenv.oneStop
    padSize = cmdenv.padSize
    planetary = cmdenv.planetary
    fleet = cmdenv.fleet
    odyssey = cmdenv.odyssey
    wantNoPlanet = cmdenv.noPlanet
    wantBlackMarket = cmdenv.blackMarket
    mls = cmdenv.maxLs
    
    stations = defaultdict(list)
    stationByID = tdb.stationByID
    
    cur = sql_query(cmdenv, tdb, queries, mode)
    for (ID, stationID, price, units) in cur:
        station = stationByID[stationID]
        if padSize and not station.checkPadSize(padSize):
            continue
        if planetary and not station.checkPlanetary(planetary):
            continue
        if fleet and not station.checkFleet(fleet):
            continue
        if odyssey and not station.checkOdyssey(odyssey):
            continue
        if wantNoPlanet and station.planetary != 'N':
            continue
        if wantBlackMarket and station.blackMarket != 'Y':
            continue
        if station in avoidStations:
            continue
        if station.system in avoidSystems:
            continue
        maxAge, stnAge = cmdenv.maxAge, station.dataAge or float("inf")
        if maxAge and stnAge > maxAge:
            continue
        
        row = ResultRow()
        row.station = station
        if mls:
            distanceFromStar = station.lsFromStar
            if distanceFromStar > mls:
                continue
        if distanceFn:
            distance = distanceFn(row.station.system)
            if distance > maxLy:
                continue
            row.dist = distance
        row.item = queries[ID]
        row.price = price
        row.units = units
        row.age = station.itemDataAgeStr
        if oneStopMode:
            stationRows = stations[stationID]
            stationRows.append(row)
            if len(stationRows) >= len(queries):
                results.rows.extend(stationRows)
        else:
            results.rows.append(row)
    
    if not results.rows:
        if oneStopMode and len(stations):
            raise NoDataError("No one-stop stations found")
        raise NoDataError("No available items found")
    
    if oneStopMode and not singleMode:
        results.rows.sort(key = lambda result: result.item.name())
    results.rows.sort(key = lambda result: result.station.name())
    if cmdenv.sortByUnits:
        results.summary.sort = "units"
        results.rows.sort(key = lambda result: result.price)
        results.rows.sort(key = lambda result: result.units, reverse = True)
    else:
        if not oneStopMode:
            results.summary.sort = "Price"
            results.rows.sort(key = lambda result: result.units, reverse = True)
            results.rows.sort(key = lambda result: result.price)
        if nearSystem and not cmdenv.sortByPrice:
            results.summary.sort = "Ly"
            results.rows.sort(key = lambda result: result.dist)
    
    limit = cmdenv.limit or 0
    if limit > 0:
        results.rows = results.rows[:limit]
    
    return results



#######################################################################
# # Transform result set into output


def render(results, cmdenv, tdb):
    mode = results.summary.mode
    singleMode = len(results.summary.queries) == 1
    maxStnLen = max_len(results.rows, key = lambda row: row.station.name())
    
    stnRowFmt = RowFormat()
    stnRowFmt.addColumn('Station', '<', maxStnLen,
            key = lambda row: row.station.name())
    if not singleMode:
        maxItmLen = max_len(results.rows, key = lambda row: row.item.name(cmdenv.detail))
        stnRowFmt.addColumn(results.summary.mode, '<', maxItmLen,
                key = lambda row: row.item.name(cmdenv.detail)
        )
    if mode is not SHIP_MODE or not singleMode:
        stnRowFmt.addColumn('Cost', '>', 10, 'n',
                key = lambda row: row.price)
    if mode is not SHIP_MODE:
        stnRowFmt.addColumn('Units', '>', 10,
                key = lambda row: '{:n}'.format(row.units) if row.units >= 0 else '?')
    
    if cmdenv.nearSystem:
        stnRowFmt.addColumn('DistLy', '>', 6, '.2f',
                key = lambda row: row.dist)
    
    if mode is not SHIP_MODE:
        stnRowFmt.addColumn('Age/days', '>', 7,
                key = lambda row: row.age)
    stnRowFmt.addColumn("StnLs", '>', 10,
            key = lambda row: row.station.distFromStar())
    stnRowFmt.addColumn('B/mkt', '>', 4,
            key = lambda row: TradeDB.marketStates[row.station.blackMarket])
    stnRowFmt.addColumn("Pad", '>', '3',
            key = lambda row: TradeDB.padSizes[row.station.maxPadSize])
    stnRowFmt.addColumn("Plt", '>', '3',
            key = lambda row: TradeDB.planetStates[row.station.planetary])
    stnRowFmt.addColumn("Flc", '>', '3',
            key = lambda row: TradeDB.fleetStates[row.station.fleet])
    stnRowFmt.addColumn("Ody", '>', '3',
            key = lambda row: TradeDB.odysseyStates[row.station.odyssey])
    
    if not cmdenv.quiet:
        heading, underline = stnRowFmt.heading()
        print(heading, underline, sep = '\n')
    
    for row in results.rows:
        print(stnRowFmt.format(row))
    
    if singleMode and cmdenv.detail:
        msg = "-- Ship Cost" if mode is SHIP_MODE else "-- Average"
        print(f"{msg:{maxStnLen}} {results.summary.avg:>10n}")



import random

import pytest

from tradedangerous.tradedb import TradeDB, Station, System
from .helpers import copy_fixtures

ORIGIN = 'Shinrarta Dezhra'
# tdb = None

def setup_module():
    copy_fixtures()

def route_to_closest(tdb: TradeDB, origin, destinations, maxLy=15):
    closest = min(destinations, key=lambda candidate: candidate.distanceTo(origin))
    print("Closest:", closest.name(), closest.distanceTo(origin))
    route = tdb.getRoute(origin, closest, maxLy)
    if not route:
        print("No route found.")
    else:
        print("Route:", ", ".join(system.name() for system, distance in route))
    return route

def should_skip() -> bool:
    return False  # os.getenv("CI") != None


class TestPeek:
    """
    Tests based on https://github.com/eyeonus/Trade-Dangerous/wiki/Python-Quick-Peek
    """
    
    @pytest.mark.skipif(should_skip(), reason="does not work with CI")
    def test_quick_origin(self, tdb: TradeDB):
        # Look up a particular system
        origin = tdb.lookupSystem(ORIGIN)
        
        assert 55.71875 == origin.posX
        assert 17.59375 == origin.posY
        assert 27.15625, origin.posZ
        
        stations = origin.stations
        assert len(stations) == 5
        for station in stations:
            assert isinstance(station, Station)
        
        # Look up a station
        abe1 = tdb.lookupStation("Abraham Lincoln")
        assert isinstance(abe1, Station)


        # Look up a station in a particular system
        sol = tdb.lookupSystem("sol")
        abe2 = tdb.lookupStation("Abraham Lincoln", sol)
        
        assert abe1 == abe2
    
    @pytest.mark.skipif(should_skip(), reason="does not work with CI")
    def test_quick_lookupPlace(self, tdb):
        # Look up a system or station using the flexible naming mechanism
        phoenix = tdb.lookupPlace("dunyach")
        assert str(type(phoenix)) == "<class 'tradedangerous.tradedb.Station'>"  # tell me what type of thing "phoenix" is...
        
        sol = tdb.lookupPlace("@sol")
        assert str(type(sol)) == "<class 'tradedangerous.tradedb.System'>"
        
        lave = tdb.lookupPlace("stein")
        assert isinstance(lave, System)
        
        abe = tdb.lookupPlace("sol/hamlinc")
        assert isinstance(abe, Station)
    
    @pytest.mark.skipif(should_skip(), reason="does not work with CI")
    def test_quick_five(self, tdb):
        systemTable = tdb.systemByID.values()
        visitMe = random.sample(list(systemTable), 5)
        origin = tdb.lookupPlace(ORIGIN)
        # Call distanceTo(origin) on every member of visitMe and
        # then retrieve the one with the lowest distance.
        closest = min(visitMe, key=lambda candidate: candidate.distanceTo(origin))
        print("{start} -> {dest}: {dist:.2f} ly".format(
            start=origin.name(), dest=closest.name(),
            dist=origin.distanceTo(closest),
        ))
        route = tdb.getRoute(origin, closest, 15)
        if not route:
            print("Shame, couldn't find a route.")
        else:
            # Route is a list of Systems. Turn it into a list of
            # System names...
            routeNames = [system.name() for system, distance in route]
            print("Route:", routeNames)
        
        route_to_closest(tdb, origin, visitMe)
        route_to_closest(tdb, origin, visitMe, 20)
        
        # lets change origin
        origin = tdb.lookupSystem("Toolfa")
        route_to_closest(tdb, origin, visitMe)
    
    @pytest.mark.skipif(should_skip(), reason="does not work with CI")
    def test_three_different(self, tdb):
        # lets try a different route:
        sol = tdb.lookupSystem("sol")
        lhs = tdb.lookupSystem("lhs 380")
        bhr = tdb.lookupSystem("bhritzameno")
        
        route_to_closest(tdb, sol, [lhs, bhr])
 



#!/usr/bin/env python3
# --------------------------------------------------------------------
# Copyright (C) Oliver 'kfsone' Smith 2014 <oliver@kfs.org>:
# Copyright (C) Bernd 'Gazelle' Gollesch 2016, 2017
# Copyright (C) Stefan 'Tromador' Morrell 2025
# Copyright (C) Jonathan 'eyeonus' Jones 2018 - 2025
#
# You are free to use, redistribute, or even print and eat a copy of
# this software so long as you include this copyright notice.
# I guarantee there is at least one bug neither of us knew about.
# --------------------------------------------------------------------
"""Setup for trade-dangerous"""
import sys
from setuptools import setup, find_packages

try:
    from semantic_release import setup_hook
    setup_hook(sys.argv)
except ImportError:
    pass

with open("README.md", "r") as fh:
    long_description = fh.read()

package = "tradedangerous"

exec(open("tradedangerous/version.py").read())  # pylint: disable=W0122

setup(
    name=package,
    version=__version__,  # pylint: disable=E0602
    install_requires=[
    "requests",
    "appJar",
    "ijson>=3.1",
    "rich==13.7.1",
    "importlib-metadata>=1",
    "sqlalchemy>=2.0,<3.0",
    ],
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
    packages=[
        ".",
        "tradedangerous",
        "tradedangerous.commands",
        "tradedangerous.mfd",
        "tradedangerous.mfd.saitek",
        "tradedangerous.misc",
        "tradedangerous.plugins",
        "tradedangerous.db",
    ],
    url="https://github.com/eyeonus/Trade-Dangerous",
    project_urls={
        "Bug Tracker": "https://github.com/eyeonus/Trade-Dangerous/issues",
        "Documentation": "https://github.com/eyeonus/Trade-Dangerous/wiki",
        "Source Code": "https://github.com/eyeonus/Trade-Dangerous",
    },
    author="eyeonus",
    author_email="eyeonus@gmail.com",
    description=(
        "Trade-Dangerous is a set of powerful trading tools for Elite Dangerous, "
        "organized around one of the most powerful trade run optimizers available."
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=["trade", "elite", "elite-dangerous"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Operating System :: OS Independent",
    ],
    license="MPL",
    python_requires=">=3.10",
    test_suite="tests",
    package_data={
        "tradedangerous": [
            "templates/TradeDangerous.sql",
            "templates/Added.csv",
            "templates/Category.csv",
            "templates/RareItem.csv",
            "templates/database_changes.json",
        ]
    },
    entry_points={
        "console_scripts": [
            "trade=trade:main",
            "tradegui=tradegui:main",
        ]
    },
    zip_safe=False,
)



"""
Utilities for reading from the Elite Dangerous Data Network.

Example usages:
    
    # Simple:
    import eddn
    listener = eddn.Listener()
    while True:
        batch = listener.get_batch()
        if batch:
            print("Got batch of %d" % len(batch))
    
    # Advanced:
    import eddn
    
    listener = eddn.Listener(
        minBatchTime=3,         # Allow at least 3-s for a batch,
        maxBatchTime=5,         # But allow upto 5s,
        reconnectTimeout=300,   # Reconnect after 5 minutes without data,
        burstLimit=500,         # Drain upto 500 prices between polls,
    )
    
    def handle_listener_error(e):
        print("Listener Error:", e)
    
    def process_batch(batch):
        stations = set()
        items = set()
        software = set()
        for price in batch:
            stations.add(price.station)
            items.add(price.item)
            software.add(price.software + ":" + price.version)
        print("Batch: %d entries" % len(batch))
        print("Stations: %s" % (','.join(stations)))
        print("Items: %s" % (','.join(items)))
    
    print("Listening for 100 batches")
    while listener.stats['batches'] < 100:
        batch = listener.get_batch(onerror=handle_listener_error)
        if batch:
            process_batch(batch)
        stats, errors = listener.stats, listener.errors
        if errors or (listener.stats['batches'] % 5) == 0:
            print("Stats:")
            for stat in sorted(stats.keys()):
                print("  {:<20s} {:>10n}".format(stat, stats[stat]))
        if errors:
            print("ERRORS:")
            for error in sorted(errors.keys()):
                print("  {:<20s} {:>10n}".format(error, errors[error]))
            listener.clear_errors()
    
    listener.reset_counters()
"""

# Copyright (C) Oliver 'kfsone' Smith <oliver@kfs.org> 2015
#
# Conditional permission to copy, modify, refactor or use this
# code is granted so long as attribution to the original author
# is included.

import json
import time
import zlib
import zmq

from collections import defaultdict
from collections import namedtuple


class MarketPrice(namedtuple('MarketPrice', [
        'system',
        'station',
        'item',
        'buy',
        'sell',
        'demand',
        'supply',
        'timestamp',
        'uploader',
        'software',
        'version',
        ])):
    pass


class Listener:
    """
    Provides an object that will listen to the Elite Dangerous Data Network
    firehose and capture messages for later consumption.
    
    Rather than individual updates, prices are captured across a window of
    between minBatchTime and maxBatchTime. When a new update is received,
    Rather than returning individual messages, messages are captured across
    a window of potentially several seconds and returned to the caller in
    batches.
    
    Attributes:
        zmqContext          Context this object is associated with,
        minBatchTime        Allow at least this long for a batch (ms),
        maxBatchTime        Don't allow a batch to run longer than this (ms),
        reconnectTimeout    Reconnect the socket after this long with no data,
        burstLimit          Read a maximum of this many messages between
                            timer checks
        
        subscriber          ZMQ socket we're using
        stats               Counters of nominal events
        errors              Counters of off-nominal events
        lastRecv            time of the last receive (or 0)
    """
    
    uri = 'tcp://eddn-relay.elite-markets.net:9500'
    supportedSchema = 'http://schemas.elite-markets.net/eddn/commodity/1'
    
    def __init__(
        self,
        zmqContext=None,
        minBatchTime=5.,    # seconds
        maxBatchTime=10.,   # seconds
        reconnectTimeout=180.,  # seconds
        burstLimit=200,
    ):
        assert burstLimit > 0
        if not zmqContext:
            zmqContext = zmq.Context()
        self.zmqContext = zmqContext
        self.subscriber = None
        
        self.minBatchTime = minBatchTime
        self.maxBatchTime = maxBatchTime
        self.reconnectTimeout = reconnectTimeout
        self.burstLimit = burstLimit
        
        self.reset_counters()
        self.connect()


    def connect(self):
        """
        Start a connection
        """
        # tear up the new connection first
        if self.subscriber:
            self.subscriber.close()
            del self.subscriber
        self.subscriber = newsub = self.zmqContext.socket(zmq.SUB)
        newsub.setsockopt(zmq.SUBSCRIBE, b"")
        newsub.connect(self.uri)
        self.lastRecv = time.time()
        self.lastJsData = None


    def disconnect(self):
        del self.subscriber


    def clear_errors(self):
        self.errors = defaultdict(int)


    def reset_counters(self):
        self.clear_errors()
        self.stats = defaultdict(int)


    def wait_for_data(self, softCutoff, hardCutoff):
        """
        Waits for data until maxBatchTime ms has elapsed
        or cutoff (absolute time) has been reached.
        """
        
        now = time.time()
        
        cutoff = min(softCutoff, hardCutoff)
        if self.lastRecv < now - self.reconnectTimeout:
            if self.lastRecv:
                self.errors['reconnects'] += 1
            self.connect()
            now = time.time()
        
        nextCutoff = min(now + self.minBatchTime, cutoff)
        if now > nextCutoff:
            return False
        
        timeout = (nextCutoff - now) * 1000     # milliseconds
        
        # Wait for an event
        events = self.subscriber.poll(timeout=timeout)
        if events == 0:
            return False
        return True


    def get_batch(self, onerror=None):
        """
        Greedily collect deduped prices from the firehose over a
        period of between minBatchTime and maxBatchTime, with
        built-in auto-reconnection if there is nothing from the
        firehose for a period of time.
        
        As json data is decoded, it is stored in self.lastJsData.
        
        Parameters:
            onerror
                None or a function/lambda that takes an error
                string and deals with it.
        
        Returns:
            A list of MarketPrice entries based on the data read.
            Prices are deduped per System+Station+Item, so that
            if two entries are received for the same combination,
            only the most recent with the newest timestamp is kept.
        
        Errors:
            Errors are acculumated in the .errors dictionary. If you
            supply an 'onerror' function they are also passed to it.
        """
        now = time.time()
        hardCutoff = now + self.maxBatchTime
        softCutoff = now + self.minBatchTime
        
        # hoists
        supportedSchema = self.supportedSchema
        sub = self.subscriber
        stats, errors = self.stats, self.errors
        
        # Prices are stored as a dictionary of
        # (sys,stn,item) => [MarketPrice]
        # The list thing is a trick to save us having to do
        # the dictionary lookup twice.
        batch = defaultdict(list)
        
        while self.wait_for_data(softCutoff, hardCutoff):
            # When wait_for_data returns True, there is some data waiting,
            # possibly multiple messages. At this point we can afford to
            # suck down whatever is waiting in "nonblocking" mode until
            # we reach the burst limit or we get EAGAIN.
            bursts = 0
            for _ in range(self.burstLimit):
                self.lastJsData = None
                try:
                    zdata = sub.recv(flags=zmq.NOBLOCK, copy=False)
                    stats['recvs'] += 1
                except zmq.error.Again:
                    break
                
                bursts += 1
                
                try:
                    jsdata = zlib.decompress(zdata)
                except Exception as e:
                    errors['deflate'] += 1
                    if onerror:
                        onerror("zlib.decompress: %s: %s"%(type(e), e))
                    continue
                
                bdata = jsdata.decode()
                
                try:
                    data = json.loads(bdata)
                except ValueError as e:
                    errors['loads'] += 1
                    if onerror:
                        onerror("json.loads: %s: %s"%(type(e), e))
                    continue
                
                self.lastJsData = jsdata
                
                try:
                    schema = data["$schemaRef"]
                except KeyError:
                    errors['schemaref'] += 1
                    if onerror:
                        onerror("missing schema ref")
                    continue
                if schema != supportedSchema:
                    errors['schema'] += 1
                    if onerror:
                        onerror("unsupported schema: "+schema)
                    continue
                try:
                    header = data["header"]
                    message = data["message"]
                    system = message["systemName"].upper()
                    station = message["stationName"].upper()
                    item = message["itemName"].upper()
                    buy = int(message["buyPrice"])
                    sell = int(message["sellPrice"])
                    demand = message["demand"]
                    supply = message["stationStock"]
                    timestamp = message["timestamp"]
                    uploader = header["uploaderID"]
                    software = header["softwareName"]
                    swVersion = header["softwareVersion"]
                except (KeyError, ValueError) as e:
                    errors['json'] += 1
                    if onerror:
                        onerror("invalid json: %s: %s"%(type(e), e))
                    continue
                
                # We've received real data.
                stats['prices'] += 1
                
                # Normalize timestamps
                timestamp = timestamp.replace("T"," ").replace("+00:00","")
                
                # We'll get either an empty list or a list containing
                # a MarketPrice. This saves us having to do the expensive
                # index operation twice.
                oldEntryList = batch[(system, station, item)]
                if oldEntryList:
                    if oldEntryList[0].timestamp > timestamp:
                        stats['timeseq'] += 1
                        continue
                    stats['timeseq']
                else:
                    # Add a blank entry to make the list size > 0
                    oldEntryList.append(None)
                
                # Here we're replacing the contents of the list.
                # This simple array lookup is several hundred times less
                # expensive than looking up a potentially large dictionary
                # by STATION/SYSTEM:ITEM...
                oldEntryList[0] = MarketPrice(
                    system, station, item,
                    buy, sell,
                    demand, supply,
                    timestamp,
                    uploader, software, swVersion,
                )
            
            # For the edge-case where we wait 4.999 seconds and then
            # get a burst of data: stick around a little longer.
            if bursts >= self.burstLimit:
                stats['numburst'] += 1
                stats['maxburst'] = max(stats['maxburst'], bursts)
                softCutoff = min(softCutoff, time.time() + 0.5)
        
        # to get average batch length, divide batchlen/batches.
        # you could do the same with prices/batches except that
        stats['batches'] += 1
        if not batch:
            stats['emptybatches'] += 1
        else:
            stats['batchlen'] += len(batch)
        
        return [ entry[0] for entry in batch.values() ]



# The runtime environment TD tools are expected to run with is encapsulated
# into a single object, the TradeEnv. See TradeEnv docstring for more.
from __future__ import annotations

from pathlib import Path
import os
import sys
import traceback
import typing

# Import some utilities from the 'rich' library that provide ways to colorize and animate
# the console output, along with other useful features.
# If the user has 'EXCEPTIONS' defined to something in the environment, then we can
# immediately benefit from beautified stacktraces.
from rich.console import Console
from rich.traceback import install as install_rich_traces


if typing.TYPE_CHECKING:
    import argparse
    from typing import Any, Optional, Union


_ROOT = os.path.abspath(os.path.dirname(__file__))


# Create a single instance of the console for everyone to use, unless they really
# want to do something unusual.
CONSOLE = Console()
STDERR  = Console(stderr=True)

if os.getenv("EXCEPTIONS"):
    # This makes call stacks show additional context and do syntax highlighting
    # that can turn reading a callstack from hours into seconds.
    install_rich_traces(console=STDERR, show_locals=False, extra_lines=1)


class BaseColorTheme:
    """ A way to theme the console output colors. The default is none. """
    CLOSE:      str = ""        # code to stop the last color
    dim:        str = ""        # code to make text dim
    bold:       str = ""        # code to make text bold
    italic:     str = ""        # code to make text italic
    # blink:    NEVER = "don't you dare"
    
    # style, label
    debug, DEBUG    = dim,  "#"
    note,  NOTE     = bold, "NOTE"
    info,  INFO     = "",   "INFO"
    warn,  WARNING  = "",   "WARNING"
    
    seq_first:  str = ""        # the first item in a sequence
    seq_last:   str = ""        # the last item in a sequence
    
    # Included as examples of how you might use this to manipulate tradecal output.
    itm_units:  str = ""        # the amount of something
    itm_name:   str = ""        # name of that unit
    itm_price:  str = ""        # how much does it cost?
    
    def render(self, renderable: Any, style: str) -> str:  # pragma: no cover, pylint: disable=unused-argument
        """ Renders the given printable item with the given style; BaseColorTheme simply uses a string transformation. """
        if isinstance(renderable, str):
            return renderable  # avoid an allocation
        return str(renderable)


class BasicRichColorTheme(BaseColorTheme):
    """ Provide's 'rich' styling without our own colorization. """
    CLOSE     = "[/]"
    bold      = "[bold]"
    dim       = "[dim]"
    italic    = "[italic]"
    
    # style, label
    debug, DEBUG   = dim,  "#"
    note,  NOTE    = bold, "NOTE"
    info,  INFO    = "",   "INFO"
    warn,  WARNING = "[orange3]", "WARNING"
    
    def render(self, renderable: Any, style: str) -> str:  # pragma: no cover
        style_attr = getattr(self, style, "")
        if not style_attr:
            return renderable if isinstance(renderable, str) else str(renderable)
        return f"{style_attr}{renderable}{self.CLOSE}"


class RichColorTheme(BasicRichColorTheme):
    """ Demonstrates how you might augment the rich theme with colors to be used fin e.g tradecal. """
    DEBUG = ":spider_web:"
    NOTE  = ":information_source:"
    WARNING = ":warning:"
    INFO  = ":gear:"
    
    # e.g. First station
    seq_first = "[cyan]"
    # e.g. Last station
    seq_last  = "[blue]"
    
    # Included as examples of how you might use this to manipulate tradecal output.
    itm_units = "[yellow3]"
    itm_name  = "[yellow]"
    itm_price = "[bold]"


class BaseConsoleIOMixin:
    """ Base mixin for running output through rich. """
    console: Console
    stderr:  Console
    theme:   BaseColorTheme
    quiet:   bool
    
    def uprint(self, *args, stderr: bool = False, style: str = None, **kwargs) -> None:
        """
            unicode-safe print via console or stderr, with 'rich' markup handling.
        """
        console = self.stderr if stderr else self.console
        console.print(*args, style=style, **kwargs)


class NonUtf8ConsoleIOMixin(BaseConsoleIOMixin):
    """ Mixing for running output through rich with UTF8-translation smoothing. """
    def uprint(self, *args, stderr: bool = False, style: str = None, **kwargs) -> None:
        """ unicode-handling print: when the stdout stream is not utf-8 supporting,
            we do a little extra io work to ensure users don't get confusing unicode
            errors. When the output stream *is* utf-8.
            
            :param stderr: report to stderr instead of stdout
            :param style: specify a 'rich' console style to use when the stream supports it
        """
        console = self.stderr if stderr else self.console
        try:
            # Attempt to print; the 'file' argument isn't supported by rich, so we'll
            # need to fall-back on old print when someone specifies it.
            console.print(*args, style=style, **kwargs)
        
        except UnicodeEncodeError as e:
            # Characters in the output couldn't be translated to unicode.
            if not self.quiet:
                self.stderr.print(
                    f"{self.theme.WARN}{self.theme.bold}CAUTION: Your terminal/console couldn't handle some "
                    "text I tried to print."
                )
                if 'EXCEPTIONS' in os.environ:
                    traceback.print_exc()
                else:
                    self.stderr.print(e)
            
            # Try to translate each ary into a viable string using utf error-replacement.
            components = [
                str(arg)
                .encode(TradeEnv.encoding, errors="replace")
                .decode(TradeEnv.encoding)
                for arg in args
            ]
            console.print(*components, style=style, **kwargs)


# If the console doesn't support UTF8, use the more-complicated implementation.
if str(sys.stdout.encoding).upper() != 'UTF-8':
    Utf8SafeConsoleIOMixin = NonUtf8ConsoleIOMixin
else:
    Utf8SafeConsoleIOMixin = BaseConsoleIOMixin

ENV_DEFAULTS: dict[str, Any] = {
        'debug': 0,
        'detail': 0,
        'quiet': 0,
        'color': False,
        'theme': BaseColorTheme(),
        'persist': bool(os.environ.get('TD_PERSIST', '1')),  # Use the 'persistence' mechanimsm
        'dataDir': os.environ.get('TD_DATA') or os.path.join(os.getcwd(), 'data'),
        'csvDir': os.environ.get('TD_CSV') or os.environ.get('TD_DATA') or os.path.join(os.getcwd(), 'data'),
        'tmpDir': os.environ.get('TD_TMP') or os.path.join(os.getcwd(), 'tmp'),
        'templateDir': os.path.join(_ROOT, 'templates'),
        'cwDir': os.getcwd(),
        'console': CONSOLE,
        'stderr':  STDERR,
    }


class TradeEnv(Utf8SafeConsoleIOMixin):
    """
        TradeDangerous provides a container for runtime configuration (cli flags, etc) and io operations to
        enable normalization of things without having to pass huge sets of arguments. This includes things
        like logging and reporting functionality.
        
        To print debug lines, use DEBUG<N>, e.g. DEBUG0, which takes a format() string and parameters, e.g.
            DEBUG1("hello, {world}{}", "!", world="world")
        
        is equivalent to:
            if tdenv.debug >= 1:
                print("#hello, {world}{}".format("!", world="world"))
        
        Use "NOTE" to print remarks which can be disabled with -q.
    """
    
    debug: int
    detail: int
    color: bool
    persist: bool
    dataDir: str
    csvDir: str
    tmpDir: str
    templateDir: str
    cwDir: str
    
    encoding = sys.stdout.encoding
    
    def __init__(self, properties: dict[str, typing.Any] | argparse.Namespace | None = None, **kwargs) -> None:
        # Inject the defaults into ourselves in a dict-like way
        self.__dict__.update(ENV_DEFAULTS)
        
        # If properties is a namespace, extract the dictionary; otherwise use it as-is
        if properties and hasattr(properties, '__dict__'):  # which arparse.Namespace has
            properties = properties.__dict__
        # Merge into our dictionary
        self.__dict__.update(properties or {})
        
        # Merge the kwargs dictionary
        self.__dict__.update(kwargs or {})
        
        # When debugging has been enabled on startup, enable slightly more
        # verbose rich backtraces.
        if self.__dict__['debug']:
            install_rich_traces(console=STDERR, show_locals=True, extra_lines=2)
        
        self.theme = RichColorTheme() if self.__dict__['color'] else BasicRichColorTheme()

    @staticmethod
    def __disabled_uprint(*args, **kwargs) -> None:
        pass

    def __getattr__(self, key: str) -> Any:
        """ Return the default for attributes we don't have """
        # The first time the DEBUG attribute is referenced, register a method for it.
        disabled: bool = False
        theme_prefix: str | None = None
        theme_label:  str | None = None
        match key:
            case "WARN" if self.quiet > 1:
                disabled = True
            case "WARN":
                theme_prefix, theme_label = self.theme.warn, self.theme.WARNING
            case "NOTE" | "INFO" if self.quiet:
                disabled = True
            case "NOTE":
                theme_prefix, theme_label = self.theme.note, self.theme.NOTE
            case "INFO":
                theme_prefix, theme_label = self.theme.info, self.theme.INFO
            case _ if key.startswith("DEBUG") and self.debug <= int(key[5:]):
                disabled = True
            case _ if key.startswith("DEBUG"):
                theme_prefix, theme_label = self.theme.debug, self.theme.DEBUG

        # If there's no function but there's a theme, create a function
        if disabled:
            setattr(self, key, self.__disabled_uprint)
            return self.__disabled_uprint

        if theme_prefix is not None:
            def __log_helper(outText, *args, stderr: bool = False, **kwargs):
                try:
                    msg = str(outText) if not (args or kwargs) else str(outText).format(*args, **kwargs)
                except Exception:  # noqa  # bare exception
                    # Fallback: dump raw message + args/kwargs repr
                    msg = f"{outText} {args!r} {kwargs!r}"
                
                self.uprint(f"{theme_prefix}{theme_label}: {msg}", stderr=stderr)

            setattr(self, key, __log_helper)
            return __log_helper
        
        return None
    
    def remove_file(self, *args) -> bool:
        """ Unlinks a file, as long as it exists, and logs the action at level 1. """
        path = Path(*args)
        if not path.exists():
            return False
        path.unlink()
        self.DEBUG1(":cross_mark: deleted {}", path)
        return True
    
    def rename_file(self, *, old: os.PathLike, new: os.PathLike) -> bool:
        """
        If 'new' exists, deletes it, and then attempts to rename old -> new. If new is not specified,
        then '.old' is appended to the end of the old filename while retaining the original suffix.
        
        :param old:             The current path/name of the file.
        :param new:             The path/name to rename the file to and remove before attempting.
        :returns:               True if the file existed and was renamed.
        """
        # Promote new to a guaranteed Path and remove it if it's present.
        new = Path(new)
        self.remove_file(new)
        
        # Promote new to a guaranteed Path and confirm it exists.
        old = Path(old)
        if not old.exists():
            return False
        
        # Perform the rename and log it at level 1.
        old.rename(new)
        self.DEBUG1(":recycle: moved {} to {}", old, new)
        
        return True
